@@ -1,11 +1,13 @@
 from collections import defaultdict
 
+from bson import ObjectId, SON
+from bson.dbref import DBRef
 import pymongo
-from bson import SON
+import six
 
-from mongoengine.base.fields import UPDATE_OPERATORS
-from mongoengine.connection import get_connection
+from mongoengine.base import UPDATE_OPERATORS
 from mongoengine.common import _import_class
+from mongoengine.connection import get_connection
 from mongoengine.errors import InvalidQueryError
 from mongoengine.python_support import IS_PYMONGO_3
 
@@ -26,13 +28,13 @@ MATCH_OPERATORS = (COMPARISON_OPERATORS + GEO_OPERATORS +
                    STRING_OPERATORS + CUSTOM_OPERATORS)
 
 
+# TODO make this less complex
 def query(_doc_cls=None, **kwargs):
-    """Transform a query from Django-style format to Mongo format.
-    """
+    """Transform a query from Django-style format to Mongo format."""
     mongo_query = {}
     merge_query = defaultdict(list)
     for key, value in sorted(kwargs.items()):
-        if key == "__raw__":
+        if key == '__raw__':
             mongo_query.update(value)
             continue
 
@@ -44,8 +46,8 @@ def query(_doc_cls=None, **kwargs):
         if len(parts) > 1 and parts[-1] in MATCH_OPERATORS:
             op = parts.pop()
 
-        # Allw to escape operator-like field name by __
-        if len(parts) > 1 and parts[-1] == "":
+        # Allow to escape operator-like field name by __
+        if len(parts) > 1 and parts[-1] == '':
             parts.pop()
 
         negate = False
@@ -57,16 +59,17 @@ def query(_doc_cls=None, **kwargs):
             # Switch field names to proper names [set in Field(name='foo')]
             try:
                 fields = _doc_cls._lookup_field(parts)
-            except Exception, e:
+            except Exception as e:
                 raise InvalidQueryError(e)
             parts = []
 
             CachedReferenceField = _import_class('CachedReferenceField')
+            GenericReferenceField = _import_class('GenericReferenceField')
 
             cleaned_fields = []
             for field in fields:
                 append_field = True
-                if isinstance(field, basestring):
+                if isinstance(field, six.string_types):
                     parts.append(field)
                     append_field = False
                 # is last and CachedReferenceField
@@ -84,9 +87,9 @@ def query(_doc_cls=None, **kwargs):
             singular_ops = [None, 'ne', 'gt', 'gte', 'lt', 'lte', 'not']
             singular_ops += STRING_OPERATORS
             if op in singular_ops:
-                if isinstance(field, basestring):
+                if isinstance(field, six.string_types):
                     if (op in STRING_OPERATORS and
-                            isinstance(value, basestring)):
+                            isinstance(value, six.string_types)):
                         StringField = _import_class('StringField')
                         value = StringField.prepare_query_value(op, value)
                     else:
@@ -98,8 +101,18 @@ def query(_doc_cls=None, **kwargs):
                         value = value['_id']
 
             elif op in ('in', 'nin', 'all', 'near') and not isinstance(value, dict):
-                # 'in', 'nin' and 'all' require a list of values
-                value = [field.prepare_query_value(op, v) for v in value]
+                # Raise an error if the in/nin/all/near param is not iterable.
+                value = _prepare_query_for_iterable(field, op, value)
+
+            # If we're querying a GenericReferenceField, we need to alter the
+            # key depending on the value:
+            # * If the value is a DBRef, the key should be "field_name._ref".
+            # * If the value is an ObjectId, the key should be "field_name._ref.$id".
+            if isinstance(field, GenericReferenceField):
+                if isinstance(value, DBRef):
+                    parts[-1] += '._ref'
+                elif isinstance(value, ObjectId):
+                    parts[-1] += '._ref.$id'
 
         # if op and op not in COMPARISON_OPERATORS:
         if op:
@@ -108,15 +121,18 @@ def query(_doc_cls=None, **kwargs):
             elif op in ('match', 'elemMatch'):
                 ListField = _import_class('ListField')
                 EmbeddedDocumentField = _import_class('EmbeddedDocumentField')
-                if (isinstance(value, dict) and isinstance(field, ListField) and
-                    isinstance(field.field, EmbeddedDocumentField)):
+                if (
+                    isinstance(value, dict) and
+                    isinstance(field, ListField) and
+                    isinstance(field.field, EmbeddedDocumentField)
+                ):
                     value = query(field.field.document_type, **value)
                 else:
                     value = field.prepare_query_value(op, value)
-                value = {"$elemMatch": value}
+                value = {'$elemMatch': value}
             elif op in CUSTOM_OPERATORS:
-                NotImplementedError("Custom method '%s' has not "
-                                    "been implemented" % op)
+                NotImplementedError('Custom method "%s" has not '
+                                    'been implemented' % op)
             elif op not in STRING_OPERATORS:
                 value = {'$' + op: value}
 
@@ -125,11 +141,13 @@ def query(_doc_cls=None, **kwargs):
 
         for i, part in indices:
             parts.insert(i, part)
+
         key = '.'.join(parts)
+
         if op is None or key not in mongo_query:
             mongo_query[key] = value
         elif key in mongo_query:
-            if key in mongo_query and isinstance(mongo_query[key], dict):
+            if isinstance(mongo_query[key], dict):
                 mongo_query[key].update(value)
                 # $max/minDistance needs to come last - convert to SON
                 value_dict = mongo_query[key]
@@ -179,15 +197,16 @@ def query(_doc_cls=None, **kwargs):
 
 
 def update(_doc_cls=None, **update):
-    """Transform an update spec from Django-style format to Mongo format.
+    """Transform an update spec from Django-style format to Mongo
+    format.
     """
     mongo_update = {}
     for key, value in update.items():
-        if key == "__raw__":
+        if key == '__raw__':
             mongo_update.update(value)
             continue
         parts = key.split('__')
-        # if there is no operator, default to "set"
+        # if there is no operator, default to 'set'
         if len(parts) < 3 and parts[0] not in UPDATE_OPERATORS:
             parts.insert(0, 'set')
         # Check for an operator and transform to mongo-style if there is
@@ -201,22 +220,25 @@ def update(_doc_cls=None, **update):
                 # Support decrement by flipping a positive value's sign
                 # and using 'inc'
                 op = 'inc'
-                if value > 0:
-                    value = -value
+                value = -value
             elif op == 'add_to_set':
                 op = 'addToSet'
             elif op == 'set_on_insert':
-                op = "setOnInsert"
+                op = 'setOnInsert'
 
         match = None
         if parts[-1] in COMPARISON_OPERATORS:
             match = parts.pop()
 
+        # Allow to escape operator-like field name by __
+        if len(parts) > 1 and parts[-1] == '':
+            parts.pop()
+
         if _doc_cls:
             # Switch field names to proper names [set in Field(name='foo')]
             try:
                 fields = _doc_cls._lookup_field(parts)
-            except Exception, e:
+            except Exception as e:
                 raise InvalidQueryError(e)
             parts = []
 
@@ -224,7 +246,7 @@ def update(_doc_cls=None, **update):
             appended_sub_field = False
             for field in fields:
                 append_field = True
-                if isinstance(field, basestring):
+                if isinstance(field, six.string_types):
                     # Convert the S operator to $
                     if field == 'S':
                         field = '$'
@@ -245,11 +267,19 @@ def update(_doc_cls=None, **update):
             else:
                 field = cleaned_fields[-1]
 
-            GeoJsonBaseField = _import_class("GeoJsonBaseField")
+            GeoJsonBaseField = _import_class('GeoJsonBaseField')
             if isinstance(field, GeoJsonBaseField):
                 value = field.to_mongo(value)
 
-            if op in (None, 'set', 'push', 'pull'):
+            if op == 'pull':
+                if field.required or value is not None:
+                    if match == 'in' and not isinstance(value, dict):
+                        value = _prepare_query_for_iterable(field, op, value)
+                    else:
+                        value = field.prepare_query_value(op, value)
+            elif op == 'push' and isinstance(value, (list, tuple, set)):
+                value = [field.prepare_query_value(op, v) for v in value]
+            elif op in (None, 'set', 'push'):
                 if field.required or value is not None:
                     value = field.prepare_query_value(op, value)
             elif op in ('pushAll', 'pullAll'):
@@ -259,7 +289,7 @@ def update(_doc_cls=None, **update):
                     value = [field.prepare_query_value(op, v) for v in value]
                 elif field.required or value is not None:
                     value = field.prepare_query_value(op, value)
-            elif op == "unset":
+            elif op == 'unset':
                 value = 1
 
         if match:
@@ -269,27 +299,33 @@ def update(_doc_cls=None, **update):
         key = '.'.join(parts)
 
         if not op:
-            raise InvalidQueryError("Updates must supply an operation "
-                                    "eg: set__FIELD=value")
+            raise InvalidQueryError('Updates must supply an operation '
+                                    'eg: set__FIELD=value')
 
         if 'pull' in op and '.' in key:
             # Dot operators don't work on pull operations
             # unless they point to a list field
             # Otherwise it uses nested dict syntax
             if op == 'pullAll':
-                raise InvalidQueryError("pullAll operations only support "
-                                        "a single field depth")
+                raise InvalidQueryError('pullAll operations only support '
+                                        'a single field depth')
 
             # Look for the last list field and use dot notation until there
             field_classes = [c.__class__ for c in cleaned_fields]
             field_classes.reverse()
             ListField = _import_class('ListField')
-            if ListField in field_classes:
-                # Join all fields via dot notation to the last ListField
+            EmbeddedDocumentListField = _import_class('EmbeddedDocumentListField')
+            if ListField in field_classes or EmbeddedDocumentListField in field_classes:
+                # Join all fields via dot notation to the last ListField or EmbeddedDocumentListField
                 # Then process as normal
+                if ListField in field_classes:
+                    _check_field = ListField
+                else:
+                    _check_field = EmbeddedDocumentListField
+
                 last_listField = len(
-                    cleaned_fields) - field_classes.index(ListField)
-                key = ".".join(parts[:last_listField])
+                    cleaned_fields) - field_classes.index(_check_field)
+                key = '.'.join(parts[:last_listField])
                 parts = parts[last_listField:]
                 parts.insert(0, key)
 
@@ -297,11 +333,27 @@ def update(_doc_cls=None, **update):
             for key in parts:
                 value = {key: value}
         elif op == 'addToSet' and isinstance(value, list):
-            value = {key: {"$each": value}}
+            value = {key: {'$each': value}}
+        elif op in ('push', 'pushAll'):
+            if parts[-1].isdigit():
+                key = parts[0]
+                position = int(parts[-1])
+                # $position expects an iterable. If pushing a single value,
+                # wrap it in a list.
+                if not isinstance(value, (set, tuple, list)):
+                    value = [value]
+                value = {key: {'$each': value, '$position': position}}
+            else:
+                if op == 'pushAll':
+                    op = 'push'  # convert to non-deprecated keyword
+                    if not isinstance(value, (set, tuple, list)):
+                        value = [value]
+                    value = {key: {'$each': value}}
+                else:
+                    value = {key: value}
         else:
             value = {key: value}
         key = '$' + op
-
         if key not in mongo_update:
             mongo_update[key] = value
         elif key in mongo_update and isinstance(mongo_update[key], dict):
@@ -311,74 +363,101 @@ def update(_doc_cls=None, **update):
 
 
 def _geo_operator(field, op, value):
-    """Helper to return the query for a given geo query"""
-    if op == "max_distance":
+    """Helper to return the query for a given geo query."""
+    if op == 'max_distance':
         value = {'$maxDistance': value}
-    elif op == "min_distance":
+    elif op == 'min_distance':
         value = {'$minDistance': value}
     elif field._geo_index == pymongo.GEO2D:
-        if op == "within_distance":
+        if op == 'within_distance':
             value = {'$within': {'$center': value}}
-        elif op == "within_spherical_distance":
+        elif op == 'within_spherical_distance':
             value = {'$within': {'$centerSphere': value}}
-        elif op == "within_polygon":
+        elif op == 'within_polygon':
             value = {'$within': {'$polygon': value}}
-        elif op == "near":
+        elif op == 'near':
             value = {'$near': value}
-        elif op == "near_sphere":
+        elif op == 'near_sphere':
             value = {'$nearSphere': value}
         elif op == 'within_box':
             value = {'$within': {'$box': value}}
         else:
-            raise NotImplementedError("Geo method '%s' has not "
-                                      "been implemented for a GeoPointField" % op)
+            raise NotImplementedError('Geo method "%s" has not been '
+                                      'implemented for a GeoPointField' % op)
     else:
-        if op == "geo_within":
-            value = {"$geoWithin": _infer_geometry(value)}
-        elif op == "geo_within_box":
-            value = {"$geoWithin": {"$box": value}}
-        elif op == "geo_within_polygon":
-            value = {"$geoWithin": {"$polygon": value}}
-        elif op == "geo_within_center":
-            value = {"$geoWithin": {"$center": value}}
-        elif op == "geo_within_sphere":
-            value = {"$geoWithin": {"$centerSphere": value}}
-        elif op == "geo_intersects":
-            value = {"$geoIntersects": _infer_geometry(value)}
-        elif op == "near":
+        if op == 'geo_within':
+            value = {'$geoWithin': _infer_geometry(value)}
+        elif op == 'geo_within_box':
+            value = {'$geoWithin': {'$box': value}}
+        elif op == 'geo_within_polygon':
+            value = {'$geoWithin': {'$polygon': value}}
+        elif op == 'geo_within_center':
+            value = {'$geoWithin': {'$center': value}}
+        elif op == 'geo_within_sphere':
+            value = {'$geoWithin': {'$centerSphere': value}}
+        elif op == 'geo_intersects':
+            value = {'$geoIntersects': _infer_geometry(value)}
+        elif op == 'near':
             value = {'$near': _infer_geometry(value)}
         else:
-            raise NotImplementedError("Geo method '%s' has not "
-                                      "been implemented for a %s " % (op, field._name))
+            raise NotImplementedError(
+                'Geo method "%s" has not been implemented for a %s '
+                % (op, field._name)
+            )
     return value
 
 
 def _infer_geometry(value):
-    """Helper method that tries to infer the $geometry shape for a given value"""
+    """Helper method that tries to infer the $geometry shape for a
+    given value.
+    """
     if isinstance(value, dict):
-        if "$geometry" in value:
+        if '$geometry' in value:
             return value
         elif 'coordinates' in value and 'type' in value:
-            return {"$geometry": value}
-        raise InvalidQueryError("Invalid $geometry dictionary should have "
-                                "type and coordinates keys")
+            return {'$geometry': value}
+        raise InvalidQueryError('Invalid $geometry dictionary should have '
+                                'type and coordinates keys')
     elif isinstance(value, (list, set)):
         # TODO: shouldn't we test value[0][0][0][0] to see if it is MultiPolygon?
+        # TODO: should both TypeError and IndexError be alike interpreted?
+
         try:
             value[0][0][0]
-            return {"$geometry": {"type": "Polygon", "coordinates": value}}
-        except:
-            pass
-        try:
-            value[0][0]
-            return {"$geometry": {"type": "LineString", "coordinates": value}}
-        except:
-            pass
-        try:
-            value[0]
-            return {"$geometry": {"type": "Point", "coordinates": value}}
-        except:
+            return {'$geometry': {'type': 'Polygon', 'coordinates': value}}
+        except (TypeError, IndexError):
             pass
 
-    raise InvalidQueryError("Invalid $geometry data. Can be either a dictionary "
-                            "or (nested) lists of coordinate(s)")
+        try:
+            value[0][0]
+            return {'$geometry': {'type': 'LineString', 'coordinates': value}}
+        except (TypeError, IndexError):
+            pass
+
+        try:
+            value[0]
+            return {'$geometry': {'type': 'Point', 'coordinates': value}}
+        except (TypeError, IndexError):
+            pass
+
+    raise InvalidQueryError('Invalid $geometry data. Can be either a '
+                            'dictionary or (nested) lists of coordinate(s)')
+
+
+def _prepare_query_for_iterable(field, op, value):
+    # We need a special check for BaseDocument, because - although it's iterable - using
+    # it as such in the context of this method is most definitely a mistake.
+    BaseDocument = _import_class('BaseDocument')
+
+    if isinstance(value, BaseDocument):
+        raise TypeError("When using the `in`, `nin`, `all`, or "
+                        "`near`-operators you can\'t use a "
+                        "`Document`, you must wrap your object "
+                        "in a list (object -> [object]).")
+
+    if not hasattr(value, '__iter__'):
+        raise TypeError("The `in`, `nin`, `all`, or "
+                        "`near`-operators must be applied to an "
+                        "iterable (e.g. a list).")
+
+    return [field.prepare_query_value(op, v) for v in value]
